function [ H, gamma, LL ] = recoverHGammaExhaustively( comm, graph, coveredByOthers, partialCoverGraph )
%recoverHGammaExhaustively Returns the parameters H and gamma to describe a community 
%   Given a community, it may be a partial one, an exhaustive search is
%   performed. 
if (exist('graph', 'var') && exist('coveredByOthers', 'var'))
    globalLL = true;
else
    if (exist('graph', 'var') || exist('coveredByOthers', 'var'))
        error('Either ''graph'' and ''coveredByOthers'' need to be defined or none of them.');
    end
    globalLL = false;
end

A = comm-diag(diag(comm)); %make sure the diagonal is empty
N = length(A);
numberOfEdges=length(find(triu(A)));

% shortcut for full community
if ~globalLL && (numberOfEdges == getAreaOfTriu(N))
    H = N-1;
    gamma = N-1;
    LL = 0;
    return;
end

Hs = [];
gammas = [];
% All the edges
[I,J]=find(triu(A));

for gamma = 0:N-1
    for H = 0:gamma
        if ~isLegalParameterConfig(N,H,gamma)
            continue;
        end
        Hs(end+1) = H;
        gammas(end+1) = gamma;
    end
end
observed1sInside = nan(length(Hs),1);
for con = 1:length(Hs)
    H = Hs(con);
    gamma = gammas(con);
    
    % derived from (i+p)*(j+p) <= theta = (N-1+p)*(H+p) to stay with integers 
    gamma2NH = gamma*gamma-(N-1)*H;
    NH2gamma = N-1+H - 2*gamma;
    thetaPrime = ((N-1)*NH2gamma+gamma2NH)*(H*NH2gamma+gamma2NH);
    edgeValue=((I-1)*NH2gamma+gamma2NH).*((J-1)*NH2gamma+gamma2NH);   
    observed1sInside(con) = length(find(edgeValue <= thetaPrime));
end
%expectedAreaInside2 = arrayfun(@(H,gamma) getAreaOfExpectedCommunity(N,H,gamma, @conditionHGamma), parameterConfig(:,1), parameterConfig(:,2));
%expectedAreaInside = transpose(arrayfun(@(H,gamma) getElementsBelowHGamma(N,H,gamma), Hs, gammas));
expectedAreaInside = transpose(getElementsBelowHGammaParallel(N,Hs,gammas));
observed0sInside = expectedAreaInside - observed1sInside;
if globalLL
    availableArea = getAreaOfTriu(length(graph))- nnz(triu(coveredByOthers));
    availableOnes = length(find(triu(graph))) - nnz(triu(coveredByOthers & graph));
    [expectedAreaOutside, observed1sOutside, observed0sOutside] = globalOutsideDensity(...
             full(double(comm)), N,Hs, gammas, availableArea, availableOnes, ...
             full(double(partialCoverGraph)), expectedAreaInside, observed1sInside);
    
%       l = length(expectedAreaInside);
%       expectedAreaOutside = nan(l,1);
%       observed1sOutside = nan(l,1);
%       observed0sOutside = nan(l,1);
%       for i=1:l
%           [expectedAreaOutside(i), observed1sOutside(i), observed0sOutside(i)] = getGlobalOutsideDensity(...
%               comm, N,Hs(i), gammas(i), availableArea, availableOnes, partialCoverGraph, expectedAreaInside(i), observed1sInside(i));
%       end
%       disp(' ');
else
    expectedAreaOutside = getAreaOfTriu(length(A)) - expectedAreaInside;
    observed1sOutside = numberOfEdges - observed1sInside;
    observed0sOutside = max(0,expectedAreaOutside - observed1sOutside);
end


expect1observe1 = logMult(observed1sInside,observed1sInside./expectedAreaInside);
expect1observe0 = logMult(observed0sInside,observed0sInside./expectedAreaInside);
expect0observe1 = logMult(observed1sOutside,observed1sOutside./expectedAreaOutside);
expect0observe0 = logMult(observed0sOutside,observed0sOutside./expectedAreaOutside);

loglike =  expect1observe1 + expect1observe0 + expect0observe1 + expect0observe0;

if ~isreal(loglike)
    warning('Can''t imagine a loglikelihood! There are %i imaginary loglikelihoods', sum(imag(loglike) ~= 0));
    loglike(imag(loglike) ~= 0) = -Inf;
end

LL = max(loglike);
indicesOfEquivalentParams = find(loglike == LL);
[~,idx] = max(Hs(indicesOfEquivalentParams)); % take the simplest, i.e. the one with max H
H = Hs(indicesOfEquivalentParams(idx));
gamma = gammas(indicesOfEquivalentParams(idx));
end


function [expectedAreaOutside, observed1sOutside, observed0sOutside] = getGlobalOutsideDensity(...
    comm, N, H, gamma, availableArea, availableOnes, partialCoverGraph, ...
    expectedAreaInside, observed1sInside)
%returns expectedAreaOutside, observed1sOutside, and observed0sOutside for given H and gamma
    c = zeros(N);
    c(1:(H+1), 1:N) = 1;
    c=triu(c);
    for i=(H+1):gamma
        maxIndex = getMaxIndexHGamma(N,H,gamma,i);
        c(i+1,(i+1):maxIndex+1) = 1; 
    end
    model = triu(c);
    partialCoverGraphAndModel = nnz(partialCoverGraph & model);
    commAndPartialCoverGraphAndModel = nnz(comm & partialCoverGraph & model);
    %fprintf('*H, gamma: %i, %i --> %i, %i, \tc, model %i, %i":\n', H, gamma, partialCoverGraphAndModel, commAndPartialCoverGraphAndModel, nnz(c), nnz(model));
    expectedAreaOutside = availableArea - expectedAreaInside + partialCoverGraphAndModel;
    observed1sOutside = availableOnes - observed1sInside + commAndPartialCoverGraphAndModel;
    observed0sOutside = expectedAreaOutside - observed1sOutside;
end

function result = isLegalParameterConfig(N, H, gamma)
%isLegalParameterConfig tests if a parameter configuration is valid

result = true;

if H > gamma,
    result = false;
    return
end

% derived from constraint p >= -gamma/2 to stay with integers
divisor = ((N-1)+H-2*gamma);
if (divisor == 0) ...
    || (divisor > 0 && -gamma*divisor > 2*(gamma*gamma-(N-1)*H))...
    || (divisor < 0 && -gamma*divisor < 2*(gamma*gamma-(N-1)*H))
    result = false;
end
end

function area = getElementsBelow(N, H, gamma)
%expected number of 1s inside the community (for upper triangular only)
    N = int32(N);
    H = int32(H);
    gamma = int32(gamma);
    
    area = (H+1)*N-(H+1)-((H+1)*(H+1)-(H+1))/2;

    for i=(H+1):gamma
        % derived from maxIndex = floor(theta/(i+p)-p) for integer division
        gamma2NH = gamma*gamma-(N-1)*H;
        NH2gamma = N-1+H - 2*gamma;
        iNHgamma = i*NH2gamma+gamma2NH;
        maxIndex = idivide( (gamma-H)*(gamma-H)*(gamma-(N-1))*(gamma-(N-1))-gamma2NH*iNHgamma ,...
                            NH2gamma*iNHgamma );
        area = area + maxIndex - i;
    end
    area = double(area);
end

% via PSI function sum only until gamma
function area = getElementsBelowApprox3(N, H, gamma)
%expected number of 1s inside the community (for upper triangular only)
    N = double(N);
    H = double(H);
    gamma = double(gamma);

    gamma2NH = gamma*gamma-(N-1)*H;
     NH2gamma = N-1+H - 2*gamma;
       
     p=gamma2NH/NH2gamma;
     theta=(gamma-H)*(gamma-H)*(gamma-(N-1))*(gamma-(N-1))/(NH2gamma*NH2gamma) ;   

    fullarea = N*(H+1)+theta*(psi(gamma+1+p)-psi(H+1+p))-(gamma-H)*(p-1/2);
    %fullarea = N*(H+1)
    %for i=(H+1):(gamma)
    %    fullarea = fullarea + floor(theta/(i+p)-p) + 1
    %end
    
    area = fullarea-((gamma+1)^2+(gamma+1))/2;
end

